@@ -3,29 +3,23 @@ from microprediction.writer import MicroWriter
 from microprediction.reader import default_url
 import muid, random, time
 from collections import OrderedDict
-from microprediction.samplers import exponential_bootstrap
+from microprediction.samplers import exponential_bootstrap, approx_mode
 import pprint
 import numpy as np
-import os
-from statistics import mode, StatisticsError, median
-
-def approx_mode(xs):
-    xr = [ round(x) for x in xs]
-    try:
-        return mode(xr)
-    except StatisticsError:
-        return None
+from statistics import median
 
 
 class MicroCrawler(MicroWriter):
-    """To create a crawler just derive from this class and overwrite:
+    """To create a crawler you can derive from this class (if you want) and it is suggested you overwrite one of more of the following methods:
 
-           sample
+           sample            - Method that creates predictions
+           downtown          - Stuff to do when there is nothing terribly urgent
 
         You might also want to change
 
-           candidate_streams
-           candidate_delays
+           candidate_streams  - A list of streams you want your crawler to look at
+           candidate_delays   - A list of horizons to predict (from the set self.delays)
+           update_frequency   - How many times to predict per arriving data point
 
        Good luck!
     """
@@ -37,7 +31,9 @@ class MicroCrawler(MicroWriter):
 
 
     def candidate_streams(self):
-        """ Modify this as you see fit
+        """ Modify this as you see fit to select streams for your crawler
+            For example your crawler might like z streams or not, or might require
+            a minimum number of lags in the time series.
 
              :return [ str ]    List of stream names that your bot would be willing to participate in
 
@@ -82,7 +78,7 @@ class MicroCrawler(MicroWriter):
         return sorted(scenarios)
 
 
-    def downtime(self,seconds):
+    def downtime(self,seconds,**ignored):
         """ This will be called when a small gap opens up between times when you need to predict """
         # Be careful if you override it or you might be late for your next prediction
         time.sleep(seconds)
@@ -114,6 +110,10 @@ class MicroCrawler(MicroWriter):
         self.backoff     = dict()              # Lookup of times to sleep when feeds are likely stale
         self.message_log = list()
         self.seconds_until_next = 10           # Set just before self.sample() is called so we know how long we have
+        self.last_performance_check = time.time()
+        self.last_new_horizon = time.time()
+        self.stream_candidates = self.candidate_streams()
+        self.next_prediction_time = dict()     # Indexed by horizon
 
     def feel_like_talking(self):
         return self.quietude<=1 or random.choice(range(self.quietude))==0
@@ -183,7 +183,7 @@ class MicroCrawler(MicroWriter):
 
     def next_horizon(self, exclude=None):
         """ Choose an urgent horizon """
-        candidates = self.candidate_streams()
+        candidates = self.stream_candidates
         if exclude:
             candidates = [ c for c in candidates if not c in exclude ]
         if candidates:
@@ -198,6 +198,29 @@ class MicroCrawler(MicroWriter):
             if horizons:
                 return random.choice(horizons)
 
+    def estimate_prediction_time(self, lagged_times, delay, num_intervals):
+        # Base on lagged times of observations and the horizon, set a time at which to next create a prediction
+        if len(lagged_times) > 5:
+            dt = -approx_mode(np.diff(lagged_times))
+            if dt is None or self.sleep_time > dt:
+                dt = self.sleep_time
+            # e.g. dt=90 and delay=310  we have 80 seconds grace
+            expected_at = lagged_times[0] + dt * num_intervals
+        else:
+            expected_at = time.time() + 600
+            dt = self.sleep_time
+
+        if delay >= dt:
+            earliest = expected_at + 5
+            latest = expected_at + delay % dt - 5
+        else:
+            latest = expected_at - dt - 5
+            earliest = expected_at - 2 * dt + 5
+        if earliest > latest:
+            earliest = latest
+        predict_time = np.random.rand() * (latest - earliest) + earliest
+        return predict_time, dt, earliest, latest, expected_at
+
     def predict_and_submit(self, name, delay, lagged_times ):
         """ Given a stream and horizon, try to submit predictions """
         horizon = self.horizon_name(name=name, delay=delay)
@@ -210,7 +233,7 @@ class MicroCrawler(MicroWriter):
         if len(lagged_values or []) < self.min_lags:
             message.update({"reason": "Too few lags", "lagged_len": len(lagged_values)})
             if len(lagged_times)>5:
-                dt = (lagged_times[1]-lagged_times[-1] )/(len(lagged_times-1))
+                dt = (lagged_times[1]-lagged_times[-1] )/(len(lagged_times)-1)
                 next_predict_time = self.min_lags*dt + lagged_times[-1]
                 if next_predict_time<time.time():
                     next_predict_time = time.time()+10*60*60
@@ -232,30 +255,14 @@ class MicroCrawler(MicroWriter):
                             "values[0]": scenario_values[0],
                             "values[-1]": scenario_values[-1]})
 
-            # Update the next prediction time
-            dt = -approx_mode(np.diff(lagged_times))
-            if dt is None or self.sleep_time > dt:
-                dt = self.sleep_time
-
-            # e.g. dt=90 and delay=310  we have 80 seconds grace
             num_intervals = self.update_frequency(name=name, delay=delay)
-            expected_at = lagged_times[0] + dt * num_intervals
-            if delay>=dt:
-                earliest = expected_at + 5
-                latest   = expected_at + delay % dt - 5
-            else:
-                latest   = expected_at - dt - 5
-                earliest = expected_at - 2*dt + 5
-            if earliest>latest:
-                earliest = latest
-            next_predict_time = np.random.rand()*(latest-earliest)+earliest
+            predict_time, dt, earliest, latest, expected_at = self.estimate_prediction_time(lagged_times=lagged_times, delay=delay, num_intervals=num_intervals)
             message.update({'earliest': earliest, 'latest': latest})
             message.update({'expected_at':expected_at})
-            
-            # Check for staleness 
-            if next_predict_time>time.time():
-                message.update({'t_next':next_predict_time})
-                self.next_prediction_time[horizon] = next_predict_time
+
+            if predict_time>time.time():
+                message.update({'t_next':predict_time})
+                self.next_prediction_time[horizon] = predict_time
             else:
                 if horizon in self.backoff:
                     self.next_prediction_time[horizon] = time.time() + self.backoff[horizon]
@@ -287,36 +294,53 @@ class MicroCrawler(MicroWriter):
         # Pick up where we left off, but stagger
         self.performance = self.get_performance()
         self.active = self.get_active()
+        print('Currently predicting for ' + str(len(self.active)) + ' horizons')
+        self.stream_candidates = self.candidate_streams()
+        print('Found ' + str(len(self.stream_candidates)) + ' candidate streams.', flush=True)
         self.next_prediction_time = dict( [ (horizon, time.time() + k*self.initial_urgency_multiplier(horizon)) for k,horizon in enumerate(self.active)])
 
         print("---------- Restarting -------------")
         pprint.pprint(self.__repr__())
         print("-----------------------------------",flush=True)
 
+
         while True:
 
             # Periodically consider entering new horizons and withdrawing from others
             self.update_seconds_until_next()
             if self.seconds_until_next>np.random.randn()*120:
-                self.performance = self.get_performance()
-                self.active = self.get_active()
-                self.withdraw_from_worst()
+                if time.time()-self.last_performance_check>60*60:
+                    self.performance = self.get_performance()
+                    self.stream_candidates = self.candidate_streams()
+                    print('Found '+str(len(self.stream_candidates))+ ' candidate streams.')
+                    self.active = self.get_active()
+                    print('Currently predicting for ' + str(len(self.active)) + ' horizons')
+                    self.withdraw_from_worst()
+                    self.last_performance_check = time.time()
 
+            # If there is time consider entering a new stream, but not too often
             self.update_seconds_until_next()
             if (self.seconds_until_next>np.random.rand()*500) or (len(self.active)==0):
-                horizon = self.next_horizon(exclude=self.horizon_blacklist)
-                name, delay = self.split_horizon_name(horizon)
-                lagged_times = self.get_lagged_times(name=name)
-                self.predict_and_submit(name=name, delay=delay, lagged_times=lagged_times)
-                print(" ", )
-                print(" Found something new to predict: " + horizon)
-                print(" ",flush=True )
-                if self.feel_like_talking():
-                    print(" ")
-                    pprint.pprint(self.recent_updates())
-                    print(" ", flush=True)
+                if (time.time()-self.last_new_horizon>10*60) or (len(self.active)==0):
+                    horizon = self.next_horizon(exclude=self.horizon_blacklist)
+                    if horizon:
+                        name, delay = self.split_horizon_name(horizon)
+                        lagged_times = self.get_lagged_times(name=name)
+                        self.predict_and_submit(name=name, delay=delay, lagged_times=lagged_times)
+                        print(" ", )
+                        print(" Found something new to predict: " + horizon)
+                        print(" ",flush=True )
+                        self.last_new_horizon = time.time()
+                        if self.feel_like_talking():
+                            print(" ")
+                            pprint.pprint(self.recent_updates())
+                            print(" ", flush=True)
+                    else:
+                        print(" ")
+                        print(" Couldn't find anything else to predict, for now. ",flush=True )
+                        print(" ",flush=True)
 
-            # Maybe we chill, or do something productive like offline estimation
+            # If there is time, maybe we chill, or do something productive like offline estimation
             self.update_seconds_until_next()
             if self.seconds_until_next>5:
                 print('Downtime for '+str(self.seconds_until_next)+' seconds. Next '+ str(self.upcoming(num=1)))
@@ -337,6 +361,8 @@ class MicroCrawler(MicroWriter):
                         # Received recent data, so it is time to predict
                         name, delay = self.split_horizon_name(horizon)
                         self.predict_and_submit(name=name, delay=delay, lagged_times=lagged_times )
+                    else:
+                        time.sleep(1)
 
 
 
